/*****************************************************************************
Spartan-3 Starter Board to Parallel Port Interface
  Software interface for getting and setting signals on the parallel port
  interface.
  
  This module takes care of the low-level data exchange with the interface
  module (Parint.v) on the FPGA through the parallel port. It uses Craig
  Peacock's PortTalk driver (http://www.beyondlogic.org/porttalk/porttalk.htm)
  to access the parallel port's I/O registers under Windows XP.

  Written by Eirik Bakke, January 14, 2006
  See http://www.mit.edu/~ebakke/parint for documentation
  
******************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <malloc.h>
#include <stddef.h>
#include <stdlib.h>
#include "porttalk_user.h"
#include "parport.h"
#include "parint.h"
#include <windows.h> /* for Sleep() */

enum {
	INVALIDB = (0x100),
	INVERT = -2,
	JTAG_TMS = 0x4,
	JTAG_TDI = 0x1,
	JTAG_TCK = 0x2,
	JTAG_LPO = 0x40,
	JTAG_TDO = PAR_S_NSELECT,
};

static int initialized = 0;
static int w_in = 0, w_out = 0;
static Binval_T inputs = NULL, outputs = NULL;
static unsigned short lptadr = 0x378; /* Usual LPT1 address. */

static void setsig(int mask, int act) {
	static unsigned int prev = INVALIDB;
	assert((mask & (JTAG_TMS | JTAG_TDI | JTAG_TCK | JTAG_LPO)) == mask);

	if (prev == INVALIDB)
		prev = inp(lptadr);

	switch (act) {
		case        1:      prev = prev | mask;
		break; case 0:      prev = prev & ~mask;
		break; case INVERT: prev = prev ^ mask;
		break; default:
			assert(0);
	}
	outp(lptadr, prev);
}

static int getsig(int mask) {
	assert(mask == JTAG_TDO);
	return (inp(lptadr + 1) & mask) != 0;
}

static void clock(void) {
	setsig(JTAG_TCK, 0);
	Parint_sleep(0);
	setsig(JTAG_TCK, 1);
	Parint_sleep(0);
}

void Parint_update(void) {
	int i, j, n = max(w_in, w_out);
	assert(initialized);

	setsig(JTAG_TMS, 1);
	clock();
	setsig(JTAG_TMS, 0);

	for (i = 0, j = w_in - n; i < n; i++, j++) {
		if (i < w_out) {
			*Binval_digit(outputs, i) = getsig(JTAG_TDO);
		}
		if (j >= 0) {
			setsig(JTAG_TDI, *Binval_digit(inputs, j));
		}
		clock();
	}

	setsig(JTAG_TMS, 1);
	clock();
}

int Parint_init(int set_w_in, int set_w_out) {
	int err;
	assert(w_in >= 0 && w_out >= 0);
	assert(!initialized);

	err = OpenPortTalk() != 0;
	if (err)
		return err;

	w_in  = set_w_in;
	w_out = set_w_out;

	inputs  = Binval_new(w_in);
	outputs = Binval_new(w_out);

	outp(lptadr + 2, (
		inp(lptadr + 2) & ~(PAR_C_BIDIR | PAR_C_CNSELECT | PAR_C_CNAUTOLF)
		) | PAR_C_NINIT | PAR_C_NSTROBE);
	outp(lptadr, 0x00);

	initialized = 1;
	return 0;
}

void Parint_shutdown(void) {
	if (initialized) {
		initialized = 0;
		free(inputs);
		free(outputs);
		inputs = outputs = NULL;
		ClosePortTalk();
	}
}

void Parint_sleep(int millisecs) {
	assert(initialized);
	if (millisecs != 0)
		Sleep(millisecs);
}

Binval_T Parint_inputs(void) {
	return inputs;
}

Binval_T Parint_outputs(void) {
	return outputs;
}

void Parint_uninstall(void) {
	UninstallPortTalkDriver();
}